import java.util.*;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.IOException;

class decrypteur {
	HashMap<String, LinkedList<permutation>> MotsPossibles;
	NoeudPriorite file_cryptee;
	TreeSet<permutation> possibilites;
	double meilleurRatio;
	Lecteur lector;
	String cryp_file;
	private int constante_dfs;

	decrypteur() { 
		lector = new Lecteur();
		constante_dfs = 7;
		lector.set_constante_file(15);
	}

	public void load_dico(String dico_file) {
		lector.initialise_dico(dico_file);
	}

// demande au lecteur de charger le texte crypt en mmoire. mode = suivant que le texte est dans un fichier ou pas
	public void load_crypt(String crypt_file, int mode, permutation p) {
		cryp_file = crypt_file;
		if (mode == 1)								// fichier
			lector.initialise_crypte(crypt_file, p);
		else										// texte saisi ; crypt_file vaut alors le texte saisi
			lector.initialise_crypte2(crypt_file, p);
		MotsPossibles = lector.getMotsPossibles();
		file_cryptee = lector.getFileLecture();
	}

// prend en argument un dossier et retourne la liste des fichiers de type dictionnaire :
// commencent par dico_ et terminent par .data
	private LinkedList<String> make_list_dico(String dico_path) {
		LinkedList<String> liste_retour = new LinkedList<String>();
		File f = new File(dico_path);
		String [] files = f.list();

		for(int i =0; i < files.length; i++)
			if (new File(dico_path + "\\" + files[i]).isFile() && files[i].length() > 5 && files[i].substring(0,5).equals("dico_") && files[i].substring(files[i].length() - 5).equals(".data"))
				liste_retour.add(dico_path + "\\" + files[i]);
		return liste_retour;
	}


// la fonction conseille dans le dm
	public void analyse_brutale(String crypt_file, int mode) {
		permutation sigma = new permutation();
		load_crypt(crypt_file, mode, sigma); // fichier  dchiffrer, mode de lecture
// sigma : utile pour une analyse frquence, vide ici normalement, c'est la permutation de base que les mots doivent tous accepter
		possibilites = new TreeSet<permutation> (new comparateur()); // stockage des possibilits, apres parcours dfs
		meilleurRatio = 0.0;
		parcours_dfs(sigma, file_cryptee);
	}

// plus fin que analyse brutale : recherche un dbut de permutation grce aux frquences
// fonctionne pour le francais
// critre : e 14,7 puis s et a 8 puis i 7,3 puis n et t 7,1 (dpend des sites sur lesquels on regarde)
	public void analyse_frequence(String crypt_file, int mode) {
		permutation sigma = new permutation();
		int[] t = new int[26];
		for(NoeudPriorite no = file_cryptee; no != null; no = no.suiv)
			for(int k = 0; k < no.s.length(); k++)
			t[no.s.charAt(k) - 'a']++;
		Arrays.sort(t);

		sigma.ajoute(t[25], 'e' - 'a');
//		sigma.ajoute(t[24], 's' - 'a');
//		sigma.ajoute(t[23], 0); // 0 <-> A
// le paramtre sigma de load_crypt est utilis ici : c'est la permutation de base que les mots doivent tous accepter
// les listes de mots possibles tiennent compte de sigma (sigma est "vide" normalement)
		load_crypt(crypt_file, mode, sigma);
		possibilites = new TreeSet<permutation> (new comparateur());
		meilleurRatio = 0.0;
		parcours_dfs(sigma, file_cryptee);
	}

// le parcours de l'arbre des permutations possibles, en profondeur d'abord, et  chaque tape on
// enrichit sa connaissance de l'information sur la permutation
	private void parcours_dfs(permutation sigma, NoeudPriorite node) {
		// critere d'arret
		if (sigma.get_nb_parcours() > constante_dfs && sigma.ratio() < .8) return;
		if (sigma.get_nb_parcours() > (2 * constante_dfs) && sigma.ratio() < (meilleurRatio * .95)) return;

		// si la permutation est dja calcule entirement, inutile de chercher  la complter
		//Cependant, on ne vrifie pas si les autres mots sont valides
		if (sigma.nb_lettres == lector.nb_lettres_utilisees)
		{
			sigma.incr_nb_parcours();
			String mot = sigma.to_mot(node.s);
			if (lector.D_contient_mot(mot)) sigma.incr_nb_match();
			if (node.suiv == null) {possibilites.add(sigma); meilleurRatio=Math.max(meilleurRatio, sigma.ratio()); return;}
			parcours_dfs(sigma, node.suiv);
			return;
		}
		// sinon... chercher  la complter
		for(permutation p : MotsPossibles.get(node.s))
		{
			if (permutation.compatible(p, sigma)) { // compatibilit OK : on continue et on sauvegarde l'info
				permutation p_new = new permutation(p, sigma);
				p_new.set_nb_match(sigma.get_nb_match() + 1);
				p_new.set_nb_parcours(sigma.get_nb_parcours() + 1);
				if (node.suiv == null) {possibilites.add(p_new); meilleurRatio=Math.max(meilleurRatio, sigma.ratio()); return;}
				parcours_dfs(p_new, node.suiv);
			}
		}
		//Ici on traite le cas o on estime que le mot dcrypt est un nom propre : on n'amliore alors pas la permutation
		sigma.incr_nb_parcours();
		permutation p_new = sigma.duplique();
		if (node.suiv == null) {possibilites.add(p_new); meilleurRatio=Math.max(meilleurRatio, sigma.ratio()); return;}
		parcours_dfs(p_new, node.suiv);	
	}

	//Affiche les permutations dont le ratio est suprieur au pourcentage, dans un fichier texte
	// s'il n'y en a aucune, on affiche quand meme les 3 meilleures
	public void affiche_possibilites(double pourcentage) {
		try {
			FileWriter lu = new FileWriter("C:\\temporarycrypteur.txt"); // cration du fichier
			BufferedWriter out = new BufferedWriter(lu);
			LinkedList<permutation> liste = new LinkedList<permutation>();
			for(permutation p : possibilites) // possibilites est un treeset. l'iterateur est par ordre croissant
				liste.addFirst(p);					// et on veut par ordre dcroissant !
			int n = 0;
			for(permutation p : liste)
				if (p.ratio() >= pourcentage || n < 3)
				try	{
					n++;
					lector.show(p, out);
				}
				catch(Exception e) {;}
			out.close();
	// ouverture de notepad
		try {
				Runtime r = Runtime.getRuntime();
				Process pr = r.exec("notepad.exe C:\\temporarycrypteur.txt");
				}catch(Exception e) {System.out.println("erreur d'execution du bloc note"); }
		}
		catch (IOException e) {;}
	}

	//Extension proposee par le sujet : tester de dcoder sur plusieurs dictionnaires
	public void multilingue(LinkedList<String> dictionnaires, String crypt_file, int mode)
	{
		possibilites = new TreeSet<permutation> (new comparateur());
		permutation p;
		double best_ratio = -0.;
		String best_langage = "";
		TreeSet<permutation> meilleurParLangue = new TreeSet<permutation> (new comparateur());
		for(String dictionnaire : dictionnaires)
		{
			//lecture des fichiers et du dictionnaire
			lector.initialise_dico(dictionnaire);
			permutation sigma = new permutation();
			load_crypt(crypt_file, mode, sigma);
			// initialisation des variables globales, utiles au parcours
			possibilites.clear();
			meilleurRatio=0.0;
			// parcours
			parcours_dfs(sigma, file_cryptee);
			// traitement du rsultat
			p = possibilites.last(); // la permutation au plus grand ratio (un TreeSet est rang par ordre croissant)
			if (p.ratio() > best_ratio) {best_ratio = p.ratio(); best_langage = dictionnaire;};
			meilleurParLangue.add(p);
			System.out.println("traitement du dictionnaire : " + dictionnaire);
		}
		permutation res = meilleurParLangue.last(); // la permutation au plus grand ratio
		System.out.println("Analyse multilingue terminee.\n\n");
		System.out.println("Le texte semble provenir de la langue correspondant  : " + best_langage + "\n");
		lector.show(res);
		possibilites = meilleurParLangue;
		affiche_possibilites(0.95); // sortie dans un fichier texte

	}

	public void multilingue(String dico_path, String crypt_file, int mode)
 // String dico_path : nom du dossier avec tous les noms de dictionnaires dedans
	{
		LinkedList<String> dictionnaires = make_list_dico(dico_path);
		multilingue(dictionnaires, crypt_file, mode);
	}

}