import java.util.*;
import java.lang.Math.*;

class permutation {
	final static int VIDE = -1;
	private int[] sigma;
	private int[] inv_sigma;
	private int nb_match, nb_parcours; // variables internes utiles  l'arbre de parcours
	public int nb_lettres;
	void incr_nb_match() { nb_match++;}
	void set_nb_match(int n) { nb_match = n;}
	int get_nb_match() { return nb_match;}
	void incr_nb_parcours() { nb_parcours++;}
	void set_nb_parcours(int n) { nb_parcours = n;}
	int get_nb_parcours() { return nb_parcours;}
	double ratio() { return ((double) (nb_match) / (double) (nb_parcours));}

	permutation () {
		sigma = new int[26]; // 'a' est cod par 0, ... 'z' par 25
		inv_sigma = new int[26];
		nb_lettres = 0;
		for(int k = 0; k < 26; k++)
		{
			sigma[k] = VIDE;
			inv_sigma[k]=VIDE;
		}
		nb_match = 0;
		nb_parcours = 0;
	}
	
	//	 cre la permutation associe au passage d'une chaine crypte  la chaine dcrypte.
	//	 on n'appelle cette fonction que dans le cas ou 2 mots peuvent etre associs  une telle transformation
	permutation(String crypt, String decrypt) throws Exception{
		this();
		boolean pasderreur=true;
		for(int k = 0; k < crypt.length() && pasderreur; k++)
			pasderreur = this.ajoute(crypt.charAt(k) - 'a', decrypt.charAt(k) - 'a');
		if (!pasderreur)
			throw new Exception("permutation non inversible");
	}

	permutation composer(permutation p) {
		permutation retour = new permutation();
		for(int i = 0; i < 26; i++)
			if (sigma[i] != VIDE) retour.ajoute(i, p.get(sigma[i]));
		return retour;
	}

	//	 sert  fusionner 2 permutations ; n'est appele que si les permutations sont compatibles
	permutation(permutation p1, permutation p2) {
		this();
		for(int k = 0; k < 26; k++) {
			if (p1.get(k) != VIDE) ajoute(k, p1.get(k));
			else if (p2.get(k) != VIDE) ajoute(k, p2.get(k));
				// sinon : champs initialiss  VIDE par dfaut
		}
	}
		
// dfinit sigma(i) = j pour la "progression" sur la connaissance de la permutation
// retourne faux si sigma(i) dja dfini et diffrent ou non dfini pour l'un des deux seulement, vrai sinon
	boolean ajoute(int i, int j) {
		if ((sigma[i] != VIDE) ^ (inv_sigma[j] != VIDE)) return false;
		if ((sigma[i] != VIDE) && (inv_sigma[j] != VIDE))
			return ((sigma[i] == j) && (inv_sigma[j]==i));
		sigma[i] = j;
		inv_sigma[j]=i;
		nb_lettres++;
		return true;
	}

// renvoie sigma(i) si dfini
	int get(int i) {
		return sigma[i];
	}
	
//	 renvoie inv_sigma(i) si dfini
	int get_inv(int i) {
		return inv_sigma[i];
	}

// cette fonction cre une nouvelle permutation identique  celle ci, d'implmentation machine distincte
	permutation duplique() {
		permutation bis = new permutation();
		for(int k = 0; k < 26; k++)
			bis.sigma[k] = this.sigma[k];
		bis.nb_match = this.nb_match;
		bis.nb_parcours = this.nb_parcours;
		bis.nb_lettres = this.nb_lettres;
		return bis;
	}

// teste la compatibilit de 2 permutations <=> si l'information qu'on cherche  rajouter n'apporte pas de "contradictions"
	static boolean compatible(permutation origine, permutation a_tester) {
		for(int i = 0; i < 26; i++)
		{
			if (origine.get(i) != VIDE && a_tester.get(i) != VIDE && origine.get(i) != a_tester.get(i)) return false;
			if (origine.get_inv(i) != VIDE && a_tester.get_inv(i) != VIDE && origine.get_inv(i) != a_tester.get_inv(i)) return false;
		}
		return true;
	}

	public String toString() {
		String s = "";
		for(int i = 0; i < 26; i++)
			if (sigma[i] != VIDE) s += (((char) ('a' + i)) + " -> " + ((char) ('a' + sigma[i]))) + "\t";
		return s + "\n";
	}

	//Mthode qui permet le dcryptage du mot s grce  la permutation courante
	String to_mot(String s) {
		String ret = "";
		for(int j = 0; j < s.length(); j++)
			ret += (char) (sigma[s.charAt(j) - 'a'] + 'a');
		return ret;
	}
	
	//retour la meilleure permutation de la liste l
	public static permutation meilleurePermutation(LinkedList<permutation> l) {
		permutation p=null;
		for (permutation q : l)
		{
			System.out.println(q.ratio() * 100);
			if (p==null){p=q;}
			if (p.ratio()<q.ratio()){p=q;}
		}
		return p;
	}

	public static permutation perm_aleatoire() {
		int[] t = new int[26];
		for(int k = 0; k < 26; k++)
			t[k] = k;

		for(int k = 0; k < 25; k++)
		{
			int a = k + (int)Math.floor(Math.random()*(26 - k));
			int temp = t[k];
			t[k] = t[a];
			t[a] = temp;
		};

		permutation p = new permutation();
		for(int k = 0; k < 26; k++)
			p.ajoute(k, t[k]);
		return p;
	}

}

class comparateur implements Comparator<permutation> {
	public int compare(permutation p, permutation q) {
		if (p.ratio() == q.ratio()) return 0;
		if (p.ratio() > q.ratio()) return 1;
		return (-1);
	}

	//public boolean equals(Object obj) {return false;}
}